import datetime
from copy import deepcopy
from uuid import uuid4

import pytest
from sqlalchemy.orm.exc import MultipleResultsFound

from orchestrator.db import ProductTable
from orchestrator.domain import SubscriptionModel
from orchestrator.services.subscriptions import (
    build_extended_domain_model,
    format_extended_domain_model,
    get_subscription,
    retrieve_subscription_by_subscription_instance_value,
)
from orchestrator.utils.json import json_dumps, json_loads
from test.unit_tests import fixtures

CORRECT_SUBSCRIPTION = str(uuid4())
INCORRECT_SUBSCRIPTION = str(uuid4())

subscription_mapping = {"PB_2": [{"rt_3": "info.id", "rt_2": "info2.id"}]}


def test_get_subscription_by_id(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = ProductTable.query.filter(ProductTable.name == "Product 3").one()
    subscription = fixtures.create_subscription_for_mapping(
        product, subscription_mapping, values, subscription_id=CORRECT_SUBSCRIPTION
    )
    result = get_subscription(str(subscription.subscription_id))
    assert str(result.subscription_id) == CORRECT_SUBSCRIPTION


def test_get_subscription_by_id_err(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = ProductTable.query.filter(ProductTable.name == "Product 3").one()
    fixtures.create_subscription_for_mapping(
        product, subscription_mapping, values, subscription_id=CORRECT_SUBSCRIPTION
    )

    with pytest.raises(ValueError):
        get_subscription(INCORRECT_SUBSCRIPTION)


def test_retrieve_subscription_by_subscription_instance_value_none(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = ProductTable.query.filter(ProductTable.name == "Product 3").one()
    fixtures.create_subscription_for_mapping(product, subscription_mapping, values)

    assert retrieve_subscription_by_subscription_instance_value("rt_2", "Wrong") is None


def test_retrieve_subscription_by_subscription_instance_value(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = ProductTable.query.filter(ProductTable.name == "Product 3").one()
    fixtures.create_subscription_for_mapping(
        product, subscription_mapping, values, subscription_id=CORRECT_SUBSCRIPTION
    )

    assert (
        str(retrieve_subscription_by_subscription_instance_value("rt_3", "0").subscription_id) == CORRECT_SUBSCRIPTION
    )


def test_retrieve_subscription_by_subscription_instance_value_err(generic_product_3):
    values = {"info.id": "0", "info2.id": "X"}
    product = ProductTable.query.filter(ProductTable.name == "Product 3").one()
    fixtures.create_subscription_for_mapping(product, subscription_mapping, values)
    fixtures.create_subscription_for_mapping(product, subscription_mapping, values)

    with pytest.raises(MultipleResultsFound):
        retrieve_subscription_by_subscription_instance_value("rt_3", "0")


def test_build_extended_domain_model(generic_subscription_1, generic_product_1):
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    extended_model = build_extended_domain_model(subscription)
    actual = json_loads(json_dumps(extended_model))

    # Remove dates from the result and just verify their type
    assert isinstance(actual.pop("start_date"), datetime.datetime)
    datetime.datetime.fromisoformat(actual["product"].pop("created_at"))

    assert actual == {
        "customer_descriptions": [],
        "customer_id": "2f47f65a-0911-e511-80d0-005056956c1a",
        "description": "Generic Subscription One",
        "end_date": None,
        "insync": True,
        "note": None,
        "pb_1": {
            "label": None,
            "name": "PB_1",
            "owner_subscription_id": generic_subscription_1,
            "rt_1": "Value1",
            "subscription_instance_id": str(subscription.pb_1.subscription_instance_id),
        },
        "pb_2": {
            "label": None,
            "name": "PB_2",
            "owner_subscription_id": generic_subscription_1,
            "rt_2": 42,
            "rt_3": "Value2",
            "subscription_instance_id": str(subscription.pb_2.subscription_instance_id),
        },
        "product": {
            # "created_at": "2022-12-20T10:36:36+00:00",
            "description": "Generic Product One",
            "end_date": None,
            "name": "Product 1",
            "product_id": str(generic_product_1.product_id),
            "product_type": "Generic",
            "status": "active",
            "tag": "GEN1",
        },
        # "start_date": datetime.datetime(2022, 12, 20, 10, 36, 36, tzinfo=datetime.timezone.utc),
        "status": "active",
        "subscription_id": generic_subscription_1,
    }


def test_format_extended_domain_model(generic_subscription_1, generic_product_1):
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    extended_model = build_extended_domain_model(subscription)

    # For the sake of testing, inject 2 in_use_by_ids on the subscription dict; 1 of which belongs
    # to the owner
    other_instance_id = uuid4()
    in_use_by_ids = [subscription.pb_1.subscription_instance_id, other_instance_id]
    extended_model["pb_1"]["in_use_by_ids"] = in_use_by_ids

    # Without filtering we should see the instance_id from the owner subscription
    formatted_no_filter = format_extended_domain_model(deepcopy(extended_model), filter_owner_relations=False)
    assert sorted(formatted_no_filter["pb_1"]["in_use_by_ids"]) == sorted(in_use_by_ids)

    # With filtering we should not see the instance_id from the owner subscription
    formatted_with_filter = format_extended_domain_model(deepcopy(extended_model), filter_owner_relations=True)
    assert sorted(formatted_with_filter["pb_1"]["in_use_by_ids"]) == [other_instance_id]



import itertools
import operator
from typing import Dict, List, Tuple, TypeVar

import structlog
from tabulate import tabulate

import orchestrator.workflows
from orchestrator.cli.helpers.input_helpers import _enumerate_menu_keys, _prompt_user_menu, get_user_input
from orchestrator.cli.helpers.print_helpers import COLOR, noqa_print, print_fmt, str_fmt
from orchestrator.db import ProductTable, WorkflowTable
from orchestrator.targets import Target
from orchestrator.workflows import LazyWorkflowInstance, get_workflow

# Workflows are registered via migrations with product type. For every product with the given product_type, there will be an entry in products_workflows.
# So the relationship product_type <-> workflow has cardinality 1 <-> *.  You can have zero or more workflows for a product_type, but every workflow has 1 product_type.
# This is not evident in the db schema and we need to manually maintain this relation via the code for it to be consistent.

# The workflows, products, product_blocks, product_product_blocks and products_workflows tables are all managed by code and migration files
# New workflows are added with migration files, so they can be removed with migration files as well
# We can assume that production environments are always test envs in terms of migrations executed. It is therefore ok use workflow names in the migration files, since
# the database should only contain workflow names added from previous migration files.

# Workflows without a product associated (tasks) should not be removed

logger = structlog.get_logger(__name__)

T = TypeVar("T")


def _print_workflows_table(workflows: List[WorkflowTable]) -> None:
    items = [(wf.name, wf.target, wf.description, wf.products[0].product_type) for wf in workflows if wf.products]
    print_fmt(
        tabulate(
            items,
            headers=["#", "name", "target", "description", "product_type"],
            showindex=itertools.count(1),
        ),
        end="\n\n",
    )


def _add_workflow(workflows: Dict[str, LazyWorkflowInstance], state: dict) -> dict:
    print_fmt("\nAdd new workflow\n", flags=[COLOR.UNDERLINE])

    if not workflows:
        noqa_print("No registered workflows found to add to the database")
        return state

    registered_product_types = [
        row[0] for row in ProductTable.query.with_entities(ProductTable.product_type).distinct()
    ]
    if not registered_product_types:
        noqa_print("No registered product types found")
        return state

    noqa_print("Which product type should the workflow be added to?")
    product_type = _prompt_user_menu(
        [*[(p, p) for p in registered_product_types], ("cancel", None)],
        keys=[*_enumerate_menu_keys(registered_product_types), "q"],
    )

    if not product_type:
        # Menu cancelled
        return state

    noqa_print(f"\nAdd product type {str_fmt(product_type, flags=[COLOR.BOLD])} to which workflow?")

    already_used_workflows = {wf["name"] for wf in state["workflows_to_add"] + state["workflows_to_delete"]}
    wf_options = [(wf, wf) for wf in workflows if wf not in already_used_workflows]
    wf_name = _prompt_user_menu([*wf_options, ("cancel", None)], keys=[*_enumerate_menu_keys(wf_options), "q"])
    if not wf_name:
        # Menu cancelled
        return state

    wf_inst = get_workflow(wf_name)
    if wf_inst is None:
        # Error getting workflow
        noqa_print("Could not load workflow")
        return state

    wf_target = wf_inst.target.value if wf_inst.target is not None else None
    wf_description = wf_inst.description
    wf_to_add = {"name": wf_name, "target": wf_target, "description": wf_description, "product_type": product_type}
    return {**state, "workflows_to_add": [*state["workflows_to_add"], wf_to_add]}


def _delete_workflow(workflows: List[WorkflowTable], state: dict) -> dict:
    print_fmt("\nDelete existing workflow\n", flags=[COLOR.UNDERLINE])
    already_used_workflows = {wf["name"] for wf in state["workflows_to_add"] + state["workflows_to_delete"]}
    items = [
        (wf.name, wf.target, wf.description, wf.products[0].product_type)
        for wf in workflows
        if wf.products and wf.name not in already_used_workflows
    ]
    items = sorted(items, key=operator.itemgetter(3, 1))
    keys = ["#", "name", "target", "description", "product_type"]
    if not items:
        noqa_print("No deletable workflows in database")
        return state

    print_fmt(
        tabulate(
            items,
            headers=keys,
            showindex=itertools.count(1),
        ),
        end="\n\n",
    )
    wf_num = get_user_input("Which workflow do you want to delete? (q to cancel) ", "q")
    if not wf_num.isdigit():
        return state
    wf_index = int(wf_num) - 1
    if 0 <= wf_index < len(items):
        item = dict(zip(keys[1:], items[wf_index]))
        return {**state, "workflows_to_delete": [*state["workflows_to_delete"], item]}
    else:
        return state


def _show_state(state: dict) -> dict:
    print_fmt("\nWorkflows to add:", flags=[COLOR.GREEN])
    print_fmt(
        tabulate(
            state["workflows_to_add"],
            headers="keys",
            showindex=itertools.count(1),
        ),
        end="\n\n",
    )

    print_fmt("Workflows to delete:", flags=[COLOR.RED])
    print_fmt(
        tabulate(
            state["workflows_to_delete"],
            headers="keys",
            showindex=itertools.count(1),
        ),
        end="\n\n",
    )
    return state


def delete_dangling_workflows(workflows: List[WorkflowTable], state: dict) -> dict:
    if not workflows:
        noqa_print("No dangling workflows found.")
        return state

    print_fmt(
        "\nThe following workflows were found in the database that do not have a corresponding LazyWorkflowInstance:\n"
    )
    # All workflow products have the same product_type, so we take the first product
    # Workflows without a product  will not be listed
    _print_workflows_table(workflows)
    should_delete_dangling_workflows = (
        get_user_input("Do you wish to delete all dangling workflows from the database? [y/n]: ", "n").lower() == "y"
    )

    if not should_delete_dangling_workflows:
        noqa_print("Cancelling")
        return state

    already_used_workflows = {wf["name"] for wf in state["workflows_to_add"] + state["workflows_to_delete"]}
    keys = ["name", "target", "description", "product_type"]
    items = [
        {
            "name": wf.name,
            "target": wf.target,
            "description": wf.description,
            "product_type": wf.products[0].product_type,
        }
        for k, wf in zip(keys, workflows)
        if wf.products and wf.name not in already_used_workflows
    ]
    return {**state, "workflows_to_delete": [*state["workflows_to_delete"], *items]}


def create_workflows_migration_wizard() -> Tuple[List[dict], List[dict]]:
    """Create tuple with lists for workflows to add and delete.

    Returns tuple:
        - list of workflow items to add in the migration
        - list of workflow items to delete in the migration
    """
    database_workflows = WorkflowTable.query
    registered_workflows = orchestrator.workflows.ALL_WORKFLOWS
    system_workflow_names = {wf.name for wf in database_workflows if wf.target == Target.SYSTEM}
    registered_non_system_workflows = {k: v for k, v in registered_workflows.items() if k not in system_workflow_names}

    unregistered_workflows = [wf for wf in database_workflows if wf.name not in registered_workflows.keys()]

    # Main menu loop
    state = {"workflows_to_add": [], "workflows_to_delete": [], "done": False}
    while not state["done"]:
        print_fmt("\nWhat do you want to do?\n", flags=[COLOR.UNDERLINE, COLOR.BOLD])
        choice_fn = _prompt_user_menu(
            [
                ("Add workflow to database", lambda s: _add_workflow(registered_non_system_workflows, s)),
                ("Delete workflow from database", lambda s: _delete_workflow(database_workflows, s)),
                (
                    "Delete unregistered workflows from database",
                    lambda s: delete_dangling_workflows(unregistered_workflows, s),
                ),
                ("Finish and create migration file", lambda s: {**s, "done": True, "abort": False}),
                ("Show current diff", _show_state),
                ("Quit menu without creating a migration file", lambda s: {**s, "done": True, "abort": True}),
            ],
            keys=["a", "x", "c", "y", "d", "q"],
        )
        if choice_fn:
            state = choice_fn(state)  # type: ignore

    if state.get("abort"):
        return [], []

    logger.info("Create workflows", create_workflows=state["workflows_to_add"])
    logger.info("Delete workflows", delete_workflows=state["workflows_to_delete"])

    return state["workflows_to_add"], state["workflows_to_delete"]  # type: ignore



import strawberry

from orchestrator.schemas.fixed_input import FixedInputConfigurationItemSchema, FixedInputSchema, TagConfig


@strawberry.experimental.pydantic.type(model=FixedInputSchema, all_fields=True)
class FixedInput:
    pass


@strawberry.experimental.pydantic.type(model=FixedInputConfigurationItemSchema, all_fields=True)
class FixedInputConfigurationItem:
    pass


@strawberry.input
class FixedInputConfigurationInputType:
    fixed_inputs: list[FixedInputConfigurationItem]
    by_tag: TagConfig



from datetime import datetime
from itertools import count
from typing import Annotated, Any, Generator, List, Optional, Union
from uuid import UUID

import strawberry
from oauth2_lib.graphql_authentication import authenticated_field
from strawberry.scalars import JSON

from orchestrator.db.models import SubscriptionTable
from orchestrator.domain.base import SubscriptionModel
from orchestrator.graphql.pagination import EMPTY_PAGE, Connection
from orchestrator.graphql.resolvers.process import resolve_processes
from orchestrator.graphql.schemas.process import ProcessType
from orchestrator.graphql.types import CustomInfo, GraphqlFilter, GraphqlSort
from orchestrator.schemas.base import OrchestratorBaseModel
from orchestrator.schemas.product import ProductSchema
from orchestrator.schemas.subscription_descriptions import SubscriptionDescriptionSchema
from orchestrator.services.subscriptions import build_extended_domain_model
from orchestrator.utils.helpers import to_camel


@strawberry.type
class SubscriptionProductBlock:
    id: int
    parent: Optional[int]
    owner_subscription_id: UUID
    resource_types: JSON


def is_product_block(candidate: Any) -> bool:
    if isinstance(candidate, dict):
        # TODO: also filter on tag (needs addition of tag in orchestrator endpoint)
        # NOTE: this crosses subscription boundaries. If needed we can add an additional filter to limit that.
        return candidate.get("owner_subscription_id", None)
    return False


def get_all_product_blocks(subscription: dict[str, Any], _tags: Optional[list[str]]) -> list[dict[str, Any]]:
    gen_id = count()

    def locate_product_block(candidate: dict[str, Any]) -> Generator:
        def new_product_block(item: dict[str, Any]) -> Generator:
            enriched_item = item | {"id": next(gen_id), "parent": candidate.get("id")}
            yield enriched_item
            yield from locate_product_block(enriched_item)

        for value in candidate.values():
            if is_product_block(value):
                yield from new_product_block(value)
            elif isinstance(value, list):
                for item in value:
                    if is_product_block(item):
                        yield from new_product_block(item)

    return list(locate_product_block(subscription))


async def get_subscription_product_blocks(
    subscription_id: UUID, tags: Optional[list[str]] = None, resource_types: Optional[list[str]] = None
) -> list[SubscriptionProductBlock]:
    subscription_model = SubscriptionModel.from_subscription(subscription_id)
    subscription = build_extended_domain_model(subscription_model)

    def to_product_block(product_block: dict[str, Any]) -> SubscriptionProductBlock:
        def is_resource_type(candidate: Any) -> bool:
            return isinstance(candidate, (bool, str, int, float, type(None)))

        def requested_resource_type(key: str) -> bool:
            return not resource_types or key in resource_types

        def included(key: str, value: Any) -> bool:
            return is_resource_type(value) and requested_resource_type(key) and key not in ("id", "parent")

        return SubscriptionProductBlock(
            id=product_block["id"],
            parent=product_block.get("parent"),
            owner_subscription_id=product_block["owner_subscription_id"],
            resource_types={to_camel(k): v for k, v in product_block.items() if included(k, v)},
        )

    product_blocks = (to_product_block(product_block) for product_block in get_all_product_blocks(subscription, tags))
    return [product_block for product_block in product_blocks if product_block.resource_types]


@strawberry.experimental.pydantic.type(model=SubscriptionDescriptionSchema, all_fields=True)
class SubscriptionDescriptionType:
    pass


class SubscriptionGraphqlSchema(OrchestratorBaseModel):
    subscription_id: UUID
    product: ProductSchema
    customer_descriptions: List[Optional[SubscriptionDescriptionSchema]] = []
    description: str
    start_date: Optional[datetime]
    end_date: Optional[datetime]
    product_id: UUID
    status: str
    insync: bool
    note: Optional[str]


@strawberry.experimental.pydantic.type(model=SubscriptionGraphqlSchema, all_fields=True)
class SubscriptionType:
    @strawberry.field(description="Return all products blocks that are part of a subscription")  # type: ignore
    async def product_blocks(
        self, tags: Optional[list[str]] = None, resource_types: Optional[list[str]] = None
    ) -> list[SubscriptionProductBlock]:
        return await get_subscription_product_blocks(self.subscription_id, tags, resource_types)  # type:ignore

    @authenticated_field(description="Returns list of processes of the subscription")  # type: ignore
    async def processes(
        self,
        info: CustomInfo,
        filter_by: Union[list[GraphqlFilter], None] = None,
        sort_by: Union[list[GraphqlSort], None] = None,
        first: int = 10,
        after: int = 0,
    ) -> Connection[ProcessType]:
        filter_by_with_related_processes = (filter_by or []) + [
            GraphqlFilter(field="subscriptionId", value=str(self.subscription_id))  # type:ignore
        ]
        return await resolve_processes(info, filter_by_with_related_processes, sort_by, first, after)

    @authenticated_field(description="Returns list of subscriptions that use this subscription")  # type: ignore
    async def in_use_by_subscriptions(
        self,
        info: CustomInfo,
        filter_by: Union[list[GraphqlFilter], None] = None,
        sort_by: Union[list[GraphqlSort], None] = None,
        first: int = 10,
        after: int = 0,
    ) -> Connection[Annotated["SubscriptionType", strawberry.lazy(".subscription")]]:
        from orchestrator.graphql.resolvers.subscription import resolve_subscriptions
        from orchestrator.services.subscriptions import query_in_use_by_subscriptions

        in_use_by_query = query_in_use_by_subscriptions(self.subscription_id)  # type:ignore
        query_results = in_use_by_query.with_entities(SubscriptionTable.subscription_id).all()
        subscription_ids = [str(s.subscription_id) for s in query_results]
        if not subscription_ids:
            return EMPTY_PAGE
        filter_by_with_related_subscriptions = (filter_by or []) + [
            GraphqlFilter(field="subscriptionIds", value=",".join(subscription_ids))
        ]
        return await resolve_subscriptions(info, filter_by_with_related_subscriptions, sort_by, first, after)

    @authenticated_field(description="Returns list of subscriptions that this subscription depends on")  # type: ignore
    async def depends_on_subscriptions(
        self,
        info: CustomInfo,
        filter_by: Union[list[GraphqlFilter], None] = None,
        sort_by: Union[list[GraphqlSort], None] = None,
        first: int = 10,
        after: int = 0,
    ) -> Connection[Annotated["SubscriptionType", strawberry.lazy(".subscription")]]:
        from orchestrator.graphql.resolvers.subscription import resolve_subscriptions
        from orchestrator.services.subscriptions import query_depends_on_subscriptions

        depends_on_query = query_depends_on_subscriptions(self.subscription_id)  # type:ignore
        query_results = depends_on_query.with_entities(SubscriptionTable.subscription_id).all()
        subscription_ids = [str(s.subscription_id) for s in query_results]
        if not subscription_ids:
            return EMPTY_PAGE
        filter_by_with_related_subscriptions = (filter_by or []) + [
            GraphqlFilter(field="subscriptionIds", value=",".join(subscription_ids))
        ]
        return await resolve_subscriptions(info, filter_by_with_related_subscriptions, sort_by, first, after)



from typing import List
from unittest import mock
from uuid import UUID

import pytest
from sqlalchemy.orm.exc import NoResultFound

from orchestrator.db import ResourceTypeTable, SubscriptionTable, WorkflowTable, db, transactional
from orchestrator.db.models import SubscriptionSearchView
from orchestrator.targets import Target


def test_transactional():
    def insert_wf(state):
        wf = WorkflowTable(name="Test transactional", target=Target.CREATE, description="Testing 1, 2, 3!")
        db.session.add(wf)

    def insert_wf_error(state):
        wf = WorkflowTable(
            name="Test transactional [ERROR]", target=Target.CREATE, description="Testing 1, 2, 3! BOOM!"
        )
        db.session.add(wf)
        raise Exception("Let's wreck some havoc!")

    logger = mock.MagicMock()

    with transactional(db, logger):
        insert_wf({})
    logger.assert_has_calls(
        [
            mock.call.debug("Temporarily disabling commit."),
            mock.call.debug("Reenabling commit."),
            mock.call.debug("Committing transaction."),
        ]
    )

    logger.reset_mock()
    with pytest.raises(Exception) as ex:
        with transactional(db, logger):
            insert_wf_error({})

    # https://github.com/PyCQA/flake8-bugbear/issues/334
    assert str(ex.value) == "Let's wreck some havoc!"

    logger.assert_has_calls(
        [
            mock.call.debug("Temporarily disabling commit."),
            mock.call.debug("Reenabling commit."),
            mock.call.warning("Rolling back transaction."),
        ]
    )


def test_transactional_no_commit():
    def insert_wf(state):
        wf = WorkflowTable(
            name="Test transactional should not be committed", target=Target.CREATE, description="Testing 1, 2, 3!"
        )
        db.session.add(wf)
        db.session.commit()

        raise Exception("Lets rollback")

    logger = mock.MagicMock()

    with pytest.raises(Exception, match="Lets rollback") as ex:
        with transactional(db, logger):
            insert_wf({})

    # https://github.com/PyCQA/flake8-bugbear/issues/334
    assert str(ex.value) == "Lets rollback"

    assert (
        db.session.query(WorkflowTable).filter(WorkflowTable.name == "Test transactional should not be committed").all()
        == []
    )
    logger.assert_has_calls(
        [
            mock.call.warning(
                "Step function tried to issue a commit. It should not! Will execute commit on behalf of step function when it returns."
            ),
        ]
    )


def test_transactional_no_commit_second_thread():
    def insert_wf(state):
        wf = WorkflowTable(
            name="Test transactional should not be committed", target=Target.CREATE, description="Testing 1, 2, 3!"
        )
        db.session.add(wf)
        db.session.commit()

        # Create new database session to simulate another workflow/api handler running at the same time
        # This is also a workaround for our disable commit wrapper, but it should be reasonable obvious that
        # someone is fucking around if you see `with db.database_scope():` in actual production code

        with db.database_scope():
            wf2 = WorkflowTable(
                name="Test transactional should be committed", target=Target.CREATE, description="Testing 1, 2, 3!"
            )
            db.session.add(wf2)
            db.session.commit()

        raise Exception("Lets rollback")

    logger = mock.MagicMock()

    with pytest.raises(Exception, match="Lets rollback") as ex:
        with transactional(db, logger):
            insert_wf({})

    # https://github.com/PyCQA/flake8-bugbear/issues/334
    assert str(ex.value) == "Lets rollback"

    assert db.session.query(WorkflowTable).filter(WorkflowTable.name == "Test transactional should be committed").one()
    assert (
        db.session.query(WorkflowTable).filter(WorkflowTable.name == "Test transactional should not be committed").all()
        == []
    )
    logger.assert_has_calls(
        [
            mock.call.warning(
                "Step function tried to issue a commit. It should not! Will execute commit on behalf of step function when it returns."
            ),
        ]
    )


def test_autouse_fixture_rolls_back_aaa():
    # We want to test whether a change committed to the database in one test is visible to other tests (as in really
    # persisted to the database). Of course such a change should not be visible if our `fastapi_app` and `database`
    # autouse fixtures work as advertised.
    #
    # However, tests should be independent of each other and we cannot assume one test runs before the other. Hence
    # this test comes in two versions: one with the `_aaa` postfix and one with the `_bbb` postfix. Both will test
    # for the presence of a change the other test thinks it has committed to the database. If one of the tests (the
    # one that runs after the other) finds the change the other has committed our fixtures don't work properly.

    # Using ResourceTypeTable as it's a simple model than doesn't require foreign keys.
    rt = ResourceTypeTable(resource_type="aaa", description="aaa")
    # print(db)
    # print(dir(db))

    db.session.add(rt)
    db.session.commit()

    with pytest.raises(NoResultFound):
        ResourceTypeTable.query.filter(ResourceTypeTable.resource_type == "bbb").one()


def test_autouse_fixture_rolls_back_bbb():
    # We want to test whether a change committed to the database in one test is visible to other tests (as in really
    # persisted to the database). Of course such a change should not be visible if our `fastapi_app` and `database`
    # autouse fixtures work as advertised.
    #
    # However, tests should be independent of each other and we cannot assume one test runs before the other. Hence
    # this test comes in two versions: one with the `_aaa` postfix and one with the `_bbb` postfix. Both will test
    # for the presence of a change the other test thinks it has committed to the database. If one of the tests (the
    # one that runs after the other) finds the change the other has committed our fixtures don't work properly.

    # Using ResourceTypeTable as it's a simple model than doesn't require foreign keys.
    rt = ResourceTypeTable(resource_type="bbb", description="bbb")
    db.session.add(rt)
    db.session.commit()

    with pytest.raises(NoResultFound):
        ResourceTypeTable.query.filter(ResourceTypeTable.resource_type == "aaa").one()


@pytest.mark.skip("TODO: Implement search method on SubscriptionSearchView")
def test_full_text_search(generic_subscription_1):
    def get_subs_tsq(query: str) -> List[SubscriptionTable]:
        return SubscriptionSearchView.query.search(query).all()

    subs = get_subs_tsq("Generic Subscription One")
    assert subs[0].subscription_id == UUID(generic_subscription_1)

    subs = get_subs_tsq("description:Generic Subscription One")
    assert subs[0].subscription_id == UUID(generic_subscription_1)

    subs = get_subs_tsq("rt_2: 42")
    assert subs[0].subscription_id == UUID(generic_subscription_1)


def test_str_method():
    assert (
        str(SubscriptionTable())
        == "SubscriptionTable(subscription_id=None, description=None, status=None, product_id=None, customer_id=None, insync=None, start_date=None, end_date=None, note=None)"
    )



"""Add product_block_relations.

Revision ID: 6896a54e9483
Revises: 3c8b9185c221
Create Date: 2021-07-01 15:33:38.065653

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy_utils import UUIDType

# revision identifiers, used by Alembic.
revision = "6896a54e9483"
down_revision = "3c8b9185c221"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "product_block_relations",
        sa.Column("parent_id", UUIDType(), nullable=False),
        sa.Column("child_id", UUIDType(), nullable=False),
        sa.Column("min", sa.Integer(), nullable=True),
        sa.Column("max", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["child_id"], ["product_blocks.product_block_id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["parent_id"], ["product_blocks.product_block_id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("parent_id", "child_id"),
    )
    op.create_index("product_block_relation_p_c_ix", "product_block_relations", ["parent_id", "child_id"], unique=True)


def downgrade() -> None:
    op.drop_index("product_block_relation_p_c_ix", table_name="product_block_relations")
    op.drop_table("product_block_relations")



"""(regression)tests in relation to domain model caching."""
from http import HTTPStatus
from os import getenv

import pytest
from nwastdlib.url import URL

from orchestrator.domain.base import SubscriptionModel
from orchestrator.services.subscriptions import build_extended_domain_model
from orchestrator.utils.redis import to_redis


@pytest.mark.skipif(
    not getenv("AIOCACHE_DISABLE", "0") == "0", reason="AIOCACHE must be enabled for this test to do anything"
)
def test_cache_update_customer_description(
    test_client, generic_subscription_1, make_customer_description, cache_fixture
):
    """Check that updating subscription customer description is reflected in the cache."""
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    cust_desc = make_customer_description(
        subscription_id=subscription.subscription_id,
        customer_id=subscription.customer_id,
        description="Original description",
    )
    extended_model = build_extended_domain_model(subscription)

    # Add domainmodel to cache
    to_redis(extended_model)
    cache_fixture.extend([f"domain:{generic_subscription_1}", f"domain:etag:{generic_subscription_1}"])

    # Retrieve domain-model, customer description should be as inserted
    response1 = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert response1.json()["customer_descriptions"][0]["description"] == "Original description"

    # Update customer description
    body = {
        "id": cust_desc.id,
        "subscription_id": subscription.subscription_id,
        "customer_id": subscription.customer_id,
        "description": "Updated description",
    }
    response2 = test_client.put("/api/subscription_customer_descriptions/", json=body)
    assert HTTPStatus.NO_CONTENT == response2.status_code

    # Check that description is updated in the domain-model
    response2 = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert response2.json()["customer_descriptions"][0]["description"] == "Updated description"


@pytest.mark.skipif(
    not getenv("AIOCACHE_DISABLE", "0") == "0", reason="AIOCACHE must be enabled for this test to do anything"
)
def test_cache_delete_customer_description(
    test_client, generic_subscription_1, make_customer_description, cache_fixture
):
    """Check that deleting subscription customer description is reflected in the cache."""
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    cust_desc = make_customer_description(
        subscription_id=subscription.subscription_id,
        customer_id=subscription.customer_id,
        description="Original description",
    )
    extended_model = build_extended_domain_model(subscription)

    # Add domainmodel to cache
    to_redis(extended_model)
    cache_fixture.extend([f"domain:{generic_subscription_1}", f"domain:etag:{generic_subscription_1}"])

    # Retrieve domain-model, customer description should be as inserted
    response1 = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert response1.json()["customer_descriptions"][0]["description"] == "Original description"

    # Delete customer description
    response2 = test_client.delete(f"/api/subscription_customer_descriptions/{cust_desc.id}")
    assert HTTPStatus.NO_CONTENT == response2.status_code

    # Check that description is removed from the domain-model
    response2 = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert response2.json()["customer_descriptions"] == []


@pytest.mark.skipif(
    not getenv("AIOCACHE_DISABLE", "0") == "0", reason="AIOCACHE must be enabled for this test to do anything"
)
def test_cache_create_customer_description(
    test_client, generic_subscription_1, make_customer_description, cache_fixture
):
    """Check that creating subscription customer description is reflected in the cache."""
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    extended_model = build_extended_domain_model(subscription)

    # Add domainmodel to cache
    to_redis(extended_model)
    cache_fixture.extend([f"domain:{generic_subscription_1}", f"domain:etag:{generic_subscription_1}"])

    # Retrieve domain-model, customer description should be empty
    response1 = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert response1.json()["customer_descriptions"] == []

    # Create customer description
    body = {
        "subscription_id": subscription.subscription_id,
        "customer_id": subscription.customer_id,
        "description": "New description",
    }
    response2 = test_client.post("/api/subscription_customer_descriptions/", json=body)
    assert HTTPStatus.NO_CONTENT == response2.status_code

    # Check that description is updated in the domain-model
    response2 = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert len(response2.json()["customer_descriptions"]) > 0
    assert response2.json()["customer_descriptions"][0]["description"] == "New description"



# Copyright 2019-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from functools import partial
from typing import Callable, Optional

import pytz
from dateutil.parser import parse
from sqlalchemy import Column

from orchestrator.db.database import SearchQuery

# from sqlalchemy.sql.expression import ColumnOperators
from orchestrator.utils.helpers import to_camel

RANGE_TYPES = {
    "gt": Column.__gt__,
    "gte": Column.__ge__,
    "lt": Column.__lt__,
    "lte": Column.__le__,
    "ne": Column.__ne__,
}


def convert_to_date(value: str) -> datetime:
    """Parse iso 8601 date from string to datetime.

    Example date: "2022-07-21T03:40:48+00:00"
    """
    try:
        return parse(value).replace(tzinfo=pytz.UTC)
    except ValueError:
        raise ValueError(f"{value} is not a valid date")


def convert_to_int(value: str) -> int:
    try:
        return int(value)
    except ValueError:
        raise ValueError(f"{value} is not a valid integer")


def get_filter_value_convert_function(field: Column) -> Callable:
    if field.type.python_type == datetime:
        return convert_to_date
    if field.type.python_type == int:
        return convert_to_int
    return lambda x: x


def generic_range_filter(range_type_fn: Callable, field: Column) -> Callable[[SearchQuery, str], SearchQuery]:
    filter_operator = partial(range_type_fn, field)
    convert_filter_value = get_filter_value_convert_function(field)

    def use_filter(query: SearchQuery, value: str) -> SearchQuery:
        converted_value = convert_filter_value(value)
        return query.filter(filter_operator(converted_value))

    return use_filter


def generic_range_filters(
    column: Column, column_alias: Optional[str] = None
) -> dict[str, Callable[[SearchQuery, str], SearchQuery]]:
    column_name = to_camel(column_alias or column.name)

    return {
        f"{column_name}{operator.capitalize()}": generic_range_filter(operator_fn, column)
        for operator, operator_fn in RANGE_TYPES.items()
    }



from typing import Optional

import pytest

from orchestrator.db import ProductBlockTable, db
from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle


class ProductBlockOneNestedForTestInactive(ProductBlockModel, product_block_name="ProductBlockOneNestedForTest"):
    sub_block: Optional["ProductBlockOneNestedForTestInactive"] = None
    int_field: Optional[int] = None


class ProductBlockOneNestedForTestProvisioning(
    ProductBlockOneNestedForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
):
    sub_block: Optional["ProductBlockOneNestedForTestProvisioning"] = None
    int_field: int


class ProductBlockOneNestedForTest(ProductBlockOneNestedForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    sub_block: Optional["ProductBlockOneNestedForTest"] = None
    int_field: int


@pytest.fixture
def test_product_block_one_nested():
    # Classes defined at module level, otherwise they remain in local namespace and
    # `get_type_hints()` can't evaluate the ForwardRefs
    return ProductBlockOneNestedForTestInactive, ProductBlockOneNestedForTestProvisioning, ProductBlockOneNestedForTest


@pytest.fixture
def test_product_block_one_nested_db_in_use_by_block(resource_type_list, resource_type_int, resource_type_str):
    in_use_by_block = ProductBlockTable(
        name="ProductBlockOneNestedForTest", description="Test Block Parent", tag="TEST", status="active"
    )
    in_use_by_block.resource_types = [resource_type_int]

    db.session.add(in_use_by_block)
    db.session.commit()

    return in_use_by_block



# Copyright 2022-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from http import HTTPStatus
from typing import Any, Callable, Union

import strawberry
import structlog
from fastapi import Depends
from fastapi.routing import APIRouter
from graphql import GraphQLError, GraphQLFormattedError
from graphql.error.graphql_error import format_error
from httpx import HTTPStatusError
from oauth2_lib.graphql_authentication import authenticated_field
from starlette.requests import Request
from strawberry.fastapi import GraphQLRouter
from strawberry.http import GraphQLHTTPResponse
from strawberry.tools import merge_types
from strawberry.types import ExecutionContext, ExecutionResult
from strawberry.utils.logging import StrawberryLogger

from orchestrator.graphql.extensions.deprecation_checker_extension import make_deprecation_checker_extension
from orchestrator.graphql.extensions.error_collector_extension import ErrorCollectorExtension
from orchestrator.graphql.pagination import Connection
from orchestrator.graphql.resolvers import (
    SettingsMutation,
    resolve_processes,
    resolve_products,
    resolve_settings,
    resolve_subscriptions,
)
from orchestrator.graphql.schemas.process import ProcessType
from orchestrator.graphql.schemas.product import ProductType
from orchestrator.graphql.schemas.settings import StatusType
from orchestrator.graphql.schemas.subscription import SubscriptionType
from orchestrator.graphql.types import CustomContext
from orchestrator.security import get_oidc_user, get_opa_security_graphql
from orchestrator.settings import app_settings

api_router = APIRouter()

logger = structlog.get_logger(__name__)


@strawberry.type(description="Orchestrator queries")
class Query:
    processes: Connection[ProcessType] = authenticated_field(
        resolver=resolve_processes, description="Returns list of processes"
    )
    products: Connection[ProductType] = authenticated_field(
        resolver=resolve_products, description="Returns list of products"
    )
    subscriptions: Connection[SubscriptionType] = authenticated_field(
        resolver=resolve_subscriptions, description="Returns list of subscriptions"
    )
    settings: StatusType = authenticated_field(
        resolver=resolve_settings,
        description="Returns information about cache, workers, and global engine settings",
    )


Mutation = merge_types("Mutation", (SettingsMutation,))


class OrchestratorGraphqlRouter(GraphQLRouter):
    def __init__(self, schema: strawberry.federation.Schema, **kwargs: Any) -> None:
        super().__init__(schema, **kwargs)

    @staticmethod
    def _format_graphql_error(error: GraphQLError) -> GraphQLFormattedError:
        if isinstance(error.original_error, HTTPStatusError):
            error.extensions["http_status_code"] = {  # type: ignore
                f"{error.original_error.request.url}": error.original_error.response.status_code
            }
        return format_error(error)

    async def process_result(self, request: Request, result: ExecutionResult) -> GraphQLHTTPResponse:
        data: GraphQLHTTPResponse = {"data": result.data}

        if result.errors:
            data["errors"] = [self._format_graphql_error(error) for error in result.errors]

        return data


class OrchestratorSchema(strawberry.federation.Schema):
    def process_errors(
        self,
        errors: list[GraphQLError],
        execution_context: Union[ExecutionContext, None] = None,
    ) -> None:
        """Override error processing to reduce verbosity of the logging.

        https://strawberry.rocks/docs/types/schema#handling-execution-errors
        """
        for error in errors:
            if (
                isinstance(error.original_error, HTTPStatusError)
                and error.original_error.response.status_code == HTTPStatus.NOT_FOUND
            ):
                message = str(error.original_error).splitlines()[0]  # Strip "For more info"
                StrawberryLogger.logger.debug(message)
            else:
                StrawberryLogger.error(error, execution_context)


schema = OrchestratorSchema(
    query=Query,
    mutation=Mutation,
    enable_federation_2=app_settings.FEDEREATION_ENABLED,
    extensions=[ErrorCollectorExtension, make_deprecation_checker_extension(query=Query)],
)


def custom_context_dependency(
    get_current_user: Callable = Depends(get_oidc_user),  # noqa: B008
    get_opa_decision: Callable = Depends(get_opa_security_graphql),  # noqa: B008
) -> CustomContext:
    return CustomContext(get_current_user=get_current_user, get_opa_decision=get_opa_decision)


async def get_context(custom_context=Depends(custom_context_dependency)) -> CustomContext:  # type: ignore # noqa: B008
    return custom_context


graphql_router = OrchestratorGraphqlRouter(schema, context_getter=get_context, graphiql=app_settings.SERVE_GRAPHQL_UI)
